# This file was generated by liblab | https://liblab.com/

from __future__ import annotations
from .base import BaseModel
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .OpenAiFunctionDefinition import OpenAiFunctionDefinition


class Type_(Enum):
    FUNCTION = "function"

    def list():
        return list(map(lambda x: x.value, Type_._member_map_.values()))


class OpenAiTools(BaseModel):
    """
    OpenAI Tools
    """

    def __init__(self, type: Type_, function: OpenAiFunctionDefinition, **kwargs):
        """
        Initialize OpenAiTools
        Parameters:
        ----------
            type: str
            function: OpenAiFunctionDefinition
        """
        self.type = self._enum_matching(type, Type_.list(), "type")
        self.function = function



# This file was generated by liblab | https://liblab.com/

import re
from typing import List, Union
from enum import Enum


class BaseModel:
    """
    A base class that all models in the SDK inherit from (expect for Enum models).

    Methods
    -------
    _pattern_matching(cls, value: str, pattern: str, variable_name: str) -> str:
        Checks if a value matches a regex pattern.
        Returns the value if there's a match, otherwise throws an error.
    _enum_matching(cls, value: Union[str,Enum], enum_values: List[str], variable_name: str) -> str:
        Checks if a value (str or enum) matches the required enum values.
        Returns the value if there's a match, otherwise throws an error.
    _one_of(cls, required_array, all_array, functions_array, input_data):
        Validates whether an input_data satisfies the oneOf requirements.
    """

    def __init__(self):
        pass
    
    @classmethod
    def _pattern_matching(cls, value: str, pattern: str, variable_name: str):
        if re.match(r"{}".format(pattern), value):
            return value
        else:
            raise ValueError(f"Invalid value for {variable_name}: must match {pattern}")
    
    @classmethod
    def _enum_matching(
        cls, value: Union[str, Enum], enum_values: List[str], variable_name: str
    ):
        str_value = value.value if isinstance(value, Enum) else value
        if str_value in enum_values:
            return value
        else:
            raise ValueError(
                f"Invalid value for {variable_name}: must match one of {enum_values}"
            )

    @classmethod
    def _one_of(cls, required_array, all_array, functions_array, input_data):
        input_array = list(input_data.keys())
        for model, fields in required_array.items():
            input_copy = input_array.copy()
            matches_required = True
            for param in fields:
                if param not in input_copy:
                    matches_required = False
                    break
                input_copy.remove(param)
            if matches_required:
                matches_all = True
                for input in input_copy:
                    if input not in all_array[model]:
                        matches_all = False
                        break
                if matches_all:
                    return functions_array[model](input_data)



from typing import Iterator, Union, Dict, TypeVar
from openai.types.chat import ChatCompletionChunk, ChatCompletion, ChatCompletionMessage, ChatCompletionMessageToolCall
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_message_tool_call import Function

_T = TypeVar("_T")


class OpenAIStream():
    """OpenAIStrem"""
    def __init__(self):
        self._contents = []

    def add(self, item: _T):
        self._contents.append(item)

    def __iter__(self) -> Iterator[_T]:
        return iter(self._contents)

    def __len__(self):
        return len(self._contents)

    def __getitem__(self, index):
        return self._contents[index]


def stream_to_chat_completion(stream: Iterator[ChatCompletionChunk]) -> Union[ChatCompletion, None]:
    """OpenAI Stream to Chat Completion"""
    tools: Dict[str, ChatCompletionMessageToolCall] = {}
    chat_completion = None
    message_content = ""
    for chunk in stream:
        if not chat_completion:
            chat_completion = ChatCompletion(
                id=chunk.id,
                choices=[],
                created=chunk.created,
                model=chunk.model,
                system_fingerprint=chunk.system_fingerprint,
                object='chat.completion'
            )
        if chunk.choices and chunk.choices[0].delta and chunk.choices[0].delta:
            delta = chunk.choices[0].delta
            message_content += getattr(delta, "content", None) or ""
            if chunk.choices[0].delta.tool_calls:       
                for tool_call in chunk.choices[0].delta.tool_calls:
                    if tool_call.id:
                        tool_call_id = tool_call.id
                    if tool_call_id:
                        if tool_call_id not in tools:
                            tools[tool_call_id] = ChatCompletionMessageToolCall(
                                id=tool_call_id,
                                function=Function(arguments=getattr(tool_call.function, "arguments", None) or "",
                                name=getattr(tool_call.function, "name", None) or ""),
                                type="function"
                            )
                        else:
                            tools[tool_call_id].function.arguments += getattr(tool_call.function, "arguments", None) or ""
                            tools[tool_call_id].function.name += getattr(tool_call.function, "name", None) or ""

        if chunk.choices and chunk.choices[0].finish_reason:
            chat_completion.choices.append(
                Choice(
                    finish_reason="stop" if chunk.choices[0].finish_reason == "eos" else chunk.choices[0].finish_reason,
                    index=0,
                    logprobs=None,
                    message=ChatCompletionMessage(
                        content=None if not message_content else message_content,
                        role='assistant',
                        function_call=None,
                        tool_calls=list(tools.values())
                    ),
                )
            )
    if chat_completion and not chat_completion.choices:
        chat_completion.choices.append(
                Choice(
                    finish_reason="stop",
                    index=0,
                    logprobs=None,
                    message=ChatCompletionMessage(
                        content=None if not message_content else message_content,
                        role='assistant',
                        function_call=None
                    ),
                )
            )
    return chat_completion



# This file was generated by liblab | https://liblab.com/

from .base import BaseModel


class OpenAiFunction(BaseModel):
    """
    Represents a function call for OpenAI with an additional type attribute.
    """

    def __init__(self, name: str, arguments: str, **kwargs):
        """
        Initialize OpenAiFunction
        Parameters:
        ----------
            name: str
            arguments: str
        """
        self.name = name
        self.arguments = arguments



# This file was generated by liblab | https://liblab.com/

from urllib.parse import quote
from ..net import query_serializer
from .base import BaseService
from ..models.GetToolsRequest import GetToolsRequest as GetToolsRequestModel
from ..models.GetToolsResponse import GetToolsResponse as GetToolsResponseModel
from ..models.GetToolsResponse import (
    GetToolsResponseGuard as GetToolsResponseGuardModel,
)
from ..models.RunToolsRequest import RunToolsRequest as RunToolsRequestModel
from ..models.ToolhouseCallResults import (
    ToolhouseCallResults as ToolhouseCallResultsModel,
)


class Tools(BaseService):
    def get_tools(self, request_input: GetToolsRequestModel) -> GetToolsResponseModel:
        """
        Get Tools
        Parameters:
        ----------
            request_input: GetToolsRequestModel
        """

        url_endpoint = "/get_tools"
        headers = {"Content-Type": "application/json"}
        self._add_required_headers(headers)

        final_url = self._url_prefix + url_endpoint
        res = self._http.post(final_url, headers, request_input, True)
        if res and isinstance(res, dict):
            return GetToolsResponseGuardModel.return_one_of(res)
        return res

    def run_tools(
        self, request_input: RunToolsRequestModel
    ) -> ToolhouseCallResultsModel:
        """
        Run Tools
        Parameters:
        ----------
            request_input: RunToolsRequestModel
        """

        url_endpoint = "/run_tools"
        headers = {"Content-Type": "application/json"}
        self._add_required_headers(headers)

        final_url = self._url_prefix + url_endpoint
        res = self._http.post(final_url, headers, request_input, True)
        if res and isinstance(res, dict):
            return ToolhouseCallResultsModel(**res)
        return res



# This file was generated by liblab | https://liblab.com/

from __future__ import annotations
from .base import BaseModel
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .FunctionParameters import FunctionParameters


class OpenAiFunctionDefinition(BaseModel):
    """
    OpenAI Function Definition
    """

    def __init__(
        self,
        name: str,
        description: str = None,
        parameters: FunctionParameters = None,
        **kwargs,
    ):
        """
        Initialize OpenAiFunctionDefinition
        Parameters:
        ----------
            name: str
            description: str
            parameters: FunctionParameters
        """
        self.name = name
        self.description = description
        self.parameters = parameters



# This file was generated by liblab | https://liblab.com/

"""
Helper functions for http calls.

Functions:
    to_serialize
"""

import io
from enum import Enum


def to_serialize(obj):
    """
    Recursively converts objects into dictionaries.

    Parameters:
    ----------
        obj:
            The object to transform into a dictionary.
    """
    result = {}
    if not hasattr(obj, "__dict__") or isinstance(
        obj, (io.TextIOWrapper, io.BufferedIOBase)
    ):
        return obj
    iter_obj = obj.__dict__.items() if hasattr(obj, "__dict__") else obj.items()
    for key, value in iter_obj:
        if isinstance(value, (io.TextIOWrapper, io.BufferedIOBase)):
            result[key] = value
        elif isinstance(value, Enum):
            result[key] = value.value
        elif isinstance(value, (list, set, tuple)):
            for i in range(len(value)):
                value[i] = to_serialize(value[i])
            result[key] = value
        elif hasattr(value, "__dict__"):
            result[key] = to_serialize(value)
        else:
            result[key] = value
    return result


# response_mapper = {"type": "type_", "input": "input_"}
# request_mapper = {"type_": "type", "input_": "input"}


def rename_keys(data, mapper):
    if isinstance(data, dict):
        new_data = {}
        for key, value in data.items():
            new_key = mapper[key] if key in mapper else key
            new_data[new_key] = rename_keys(value, mapper)
        return new_data
    elif isinstance(data, list):
        return [rename_keys(item, mapper) for item in data]
    else:
        return data


def rename_reserved_keys(data):
    return data


def rename_to_reserved_keys(data):
    return data



# This file was generated by liblab | https://liblab.com/

from __future__ import annotations
from typing import Any
from .base import BaseModel
from .Provider import Provider


class GetToolsRequest(BaseModel):
    """
    Represents a tool call for Toolhouse.
    """

    def __init__(self, provider: Provider, metadata: Any):
        """
        Initialize RunToolsRequest
        Parameters:
        ----------
            provider: str
            metadata
        """
        self.provider = self._enum_matching(provider, Provider.list(), "provider")
        self.metadata = metadata



# This file was generated by liblab | https://liblab.com/

from typing import Any, Dict, List
from enum import Enum

explode = bool


def form(parameter_name: str, parameter_value: Any, explode: bool) -> str:
    if isinstance(parameter_value, Enum):
        return f"{parameter_name}=" + str(parameter_value.value)

    if isinstance(parameter_value, list):
        return (
            "&".join([f"{parameter_name}={v}" for v in parameter_value])
            if explode
            else f"{parameter_name}=" + ",".join([str(v) for v in parameter_value])
        )

    if isinstance(parameter_value, dict):
        if explode:
            # Serialize object with exploded format: "key1=value1&key2=value2"
            return "&".join([f"{k}={v}" for k, v in parameter_value.items()])
        else:
            # Serialize object with non-exploded format: "key=key1,value1,key2,value2"
            return f"{parameter_name}=" + ",".join(
                [str(item) for sublist in parameter_value.items() for item in sublist]
            )

    return f"{parameter_name}=" + str(parameter_value)


style_methods = {
    "form": form,
}


def serialize_query(parameter_style, explode, key: str, parameter_value: Any) -> str:
    method = style_methods.get(parameter_style)
    return method(key, parameter_value, explode) if method else ""



# This file was generated by liblab | https://liblab.com/
# flake8: noqa
from .Provider import Provider
from .GetToolsRequest import GetToolsRequest
from .GetToolsResponse import GetToolsResponse
from .RunToolsRequest import RunToolsRequest
from .AntropicTools import AntropicTools
from .FunctionParameters import FunctionParameters
from .OpenAiTools import OpenAiTools
from .OpenAiFunctionDefinition import OpenAiFunctionDefinition
from .AnthropicToolCall import AnthropicToolCall
from .OpenAiAssistantsToolCall import OpenAiAssistantsToolCall
from .OpenAiFunction import OpenAiFunction
from .OpenAiToolCall import OpenAiToolCall
from .ToolhouseCallResults import ToolhouseCallResults
from .AnthropicToolCallResults import AnthropicToolCallResults
from .OpenAiToolCallResults import OpenAiToolCallResults
from .OpenAiToolAssistantsCallResults import OpenAiToolAssistantsCallResults
from .OpenAIStream import OpenAIStream, stream_to_chat_completion



# This file was generated by liblab | https://liblab.com/

from __future__ import annotations
from .base import BaseModel
from typing import List
from typing import Any
from .AntropicTools import AntropicTools
from .OpenAiTools import OpenAiTools

GetToolsResponse1 = List[AntropicTools]


def returnGetToolsResponse1(input_data):
    return GetToolsResponse1(**input_data)


GetToolsResponse2 = List[OpenAiTools]


def returnGetToolsResponse2(input_data):
    return GetToolsResponse2(**input_data)


class GetToolsResponseGuard(BaseModel):
    required_lists: dict = {}
    optional_lists: dict = {}
    class_list: dict = {}

    @classmethod
    def return_one_of(cls, raw_input):
        return cls._one_of(
            cls.required_lists, cls.optional_lists, cls.class_list, raw_input
        )


GetToolsResponse = Any



# This file was generated by liblab | https://liblab.com/

from .base import BaseModel
from enum import Enum


class Type_(Enum):
    """Function _Type"""
    FUNCTION = "function"

    @classmethod
    def list(cls):
        """Return the list of Items"""
        return [member.value for member in cls]


class OpenAiAssistantsFunction(BaseModel):
    """
    Represents a function call for OpenAI with an additional type attribute.
    """

    def __init__(self,  type: Type_, name: str, arguments: str, **kwargs):
        """
        Initialize OpenAiAssistantsFunction
        Parameters:
        ----------
            type: str
            name: str
            arguments: str
        """
        self.type = self._enum_matching(type, Type_.list(), "type")
        self.name = name
        self.arguments = arguments



# This file was generated by liblab | https://liblab.com/

from enum import Enum


class Provider(Enum):
    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    OPENAI_ASSISTANTS = "openai_assistants"

    def list():
        return list(map(lambda x: x.value, Provider._member_map_.values()))



# This file was generated by liblab | https://liblab.com/

"""
An enum class containing all the possible environments that 
a user can switch between in this SDK.
"""
from enum import Enum


class Environment(Enum):
    """The environments available for this SDK"""

    DEFAULT = "https://api.toolhouse.ai/v1"



from typing import Literal, List, Dict
from pydantic import BaseModel


ArgumentType = Literal["string", "number", "integer", "object", "array", "boolean"]


class Property(BaseModel):
    """Tool Property"""
    type: ArgumentType
    description: str


class Paramenters(BaseModel):
    """Parameters"""
    type: Literal["object"]
    properties: Dict[str, Property]


class Function(BaseModel):
    """InputSchema"""
    name: str
    description: str
    parameters: Paramenters
    required: List[str]


class Tool(BaseModel):
    """OpenAI Tools"""
    type: Literal["function"]
    function: Function


class Tools(BaseModel):
    """ToolHouse"""
    tools: List[Tool]



# This file was generated by liblab | https://liblab.com/

from __future__ import annotations
from .base import BaseModel
from typing import Union
from enum import Enum
from .AnthropicToolCallResults import AnthropicToolCallResults
from .OpenAiToolCallResults import OpenAiToolCallResults
from .OpenAiToolAssistantsCallResults import OpenAiToolAssistantsCallResults


def returnAnthropicToolCallResults(input_data):
    return AnthropicToolCallResults(**input_data)


def returnOpenAiToolCallResults(input_data):
    return OpenAiToolCallResults(**input_data)


def returnOpenAiToolAssistantsCallResults(input_data):
    return OpenAiToolAssistantsCallResults(**input_data)


class ContentGuard(BaseModel):
    required_lists = {
        "AnthropicToolCallResults": ["tool_use_id", "content", "type"],
        "OpenAiToolCallResults": ["role", "tool_call_id", "name", "content"],
        "OpenAiToolAssistantsCallResults": ["tool_call_id", "output"],
    }
    optional_lists = {
        "AnthropicToolCallResults": [],
        "OpenAiToolCallResults": [],
        "OpenAiToolAssistantsCallResults": [],
    }
    class_list = {
        "AnthropicToolCallResults": returnAnthropicToolCallResults,
        "OpenAiToolCallResults": returnOpenAiToolCallResults,
        "OpenAiToolAssistantsCallResults": returnOpenAiToolAssistantsCallResults,
    }

    @classmethod
    def return_one_of(cls, raw_input):
        return cls._one_of(
            cls.required_lists, cls.optional_lists, cls.class_list, raw_input
        )


Content = Union[
    AnthropicToolCallResults, OpenAiToolCallResults, OpenAiToolAssistantsCallResults
]


class Provider(Enum):
    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    OPENAI_ASSISTANTS = "openai_assistants"

    def list():
        return list(map(lambda x: x.value, Provider._member_map_.values()))


class ToolhouseCallResults(BaseModel):
    """
    Represents the results of a tool call for Toolhouse.
    """

    def __init__(self, content: Content, provider: Provider, **kwargs):
        """
        Initialize ToolhouseCallResults
        Parameters:
        ----------
            content: Content
            provider: str
        """
        self.content = content
        self.provider = self._enum_matching(provider, Provider.list(), "provider")



from typing import Literal, List
from decimal import Decimal
from pydantic import BaseModel, Field, HttpUrl


ArgumentType = Literal["string", "number", "integer", "object", "array", "boolean"]
ArgumentFrom = Literal["llm", "user", "metadata"]
ToolType = Literal["local", "remote"]


class Argument(BaseModel):
    """Tool Arguments"""
    name: str
    type: ArgumentType
    source: ArgumentFrom
    label: str
    description: str
    required: bool


class ToolHouseTool(BaseModel):
    """ToolHouse Tools Configuration"""
    id: str
    publisher: str
    tool_type: ToolType = "remote"
    logo: str = Field(pattern=r'^data:image\/svg\+xml;base64,')
    title: str
    category: str
    short_description: str
    long_description: str
    price_per_execution: Decimal = Field(gt=0)
    star_rating: Decimal = Field(ge=0, le=5)
    executions: int
    url: HttpUrl
    description_for_model: str
    arguments: List[Argument]


class Tools(BaseModel):
    """ToolHouse"""
    tools: List[ToolHouseTool]



# This file was generated by liblab | https://liblab.com/

"""
Creates a Toolhouse class.
Generates the main SDK with all available queries as attributes.

Class:
    Toolhouse
"""
import os
from enum import Enum
from typing import List, Union, Dict, Any, Optional

from .exceptions import ToolhouseError
from .net.environment import Environment
from .services.tools import Tools
from .services.local_tools import LocalTools
from .models.Provider import Provider as ProviderModel
from .models.RunToolsRequest import RunToolsRequest
from .models.GetToolsRequest import GetToolsRequest
try:
    from .models.OpenAIStream import stream_to_chat_completion
except ImportError:
    stream_to_chat_completion = None  # type: ignore


class Toolhouse:
    """
    A class representing the full Toolhouse SDK

    Attributes
    ----------
    tools : Tools

    Methods
    -------
    set_base_url(url: str)
        Sets the end URL
    set_access_token(access_token)
        Set the access token
    """

    def __init__(self, access_token: Optional[str] = None,
                 provider: Union[ProviderModel, str] = ProviderModel.OPENAI,
                 environment: Environment = Environment.DEFAULT
                 ) -> None:
        """
         Initializes the Toolhouse SDK class.

         Parameters
         ----------
         access_token : str
             The access token
         provider : ProviderModel
             The provider model or name
         environment : Environment
             The environment that the SDK is accessing
         """
        if not provider:
            raise ValueError(
                "Parameter provider is required, cannot be empty or blank.")
        self.provider = self._enum_matching(
            provider, ProviderModel.list(), "provider")
        if access_token is None:
            access_token = os.environ.get("TOOLHOUSE_API_KEY", None)
        if access_token is None:
            raise ToolhouseError(
                "The access_token client option must be set either by passing access_token to the SDK or by setting the TOOLHOUSE_API_KEY environment variable"
            )
        self.api_key = access_token
        self.tools = Tools(access_token)
        self.metadata: Dict[str, Any] = {}
        self.set_base_url(environment.value if isinstance(
            environment, Environment) else environment)
        self.local_tools: LocalTools = LocalTools()

    def register_local_tool(self, local_tool):
        """Register Local Tools"""
        return self.local_tools.register_local_tool(local_tool)
    
    def set_metadata(self, key: str, value) -> None:
        """
        Sets User Metadata

        Parameters
        ----------
            key
            value
        """
        self.metadata[key] = value

    def set_provider(self, provider: ProviderModel) -> None:
        """
        Sets User Metadata

        Parameters
        ----------
        provider : ProviderModel
            The provider model or name
        """
        self.provider = self._enum_matching(
            provider, ProviderModel.list(), "provider")

    def set_base_url(self, url: str) -> None:
        """
        Sets the end URL

        Parameters
        ----------
            url:
                The end URL
        """
        self.tools.set_base_url(url)

    def set_access_token(self, token: str) -> None:
        """
        Sets auth token key

        Parameters
        ----------
        token: string
            Auth token value
        """
        self.tools.set_access_token(token)

    def get_tools(self):
        """
        Get Tools
        """
        return self.tools.get_tools(GetToolsRequest(provider=self.provider, metadata=self.metadata))

    def run_tools(self, response, append: bool = True, stream=False) -> List:
        """
        Run Tools based on the response.
        Parameters
        ----------
        response : Any
            The response from the provider
        
        append : bool
            Appends the LLM response to the list of messages in the return value.

        Returns
        -------
        List
            A list of messages
        """
        messages: List = []

        if self.provider in ("openai", ProviderModel.OPENAI):
            if stream:
                if stream_to_chat_completion is None:
                    raise ImportError("OpenAI Package is required for this function.")
                response = stream_to_chat_completion(response)
                if response is None:
                    return []
            if response.choices[0].finish_reason != 'tool_calls':
                return []
            response_message = response.choices[0].message
            if append:
                msg = response_message.model_dump()
                del msg["function_call"]
                messages.append(msg)
            tool_calls = getattr(response_message, 'tool_calls', None)

            if tool_calls:
                for tool in tool_calls:
                    if tool.function.name in self.local_tools.get_registered_tools():
                        result = self.local_tools.run_tools(tool)
                        messages.append(result.model_dump())
                    else:
                        run_tool_request = RunToolsRequest(
                            tool, self.provider, self.metadata)
                        run_response = self.tools.run_tools(run_tool_request)
                        messages.append(run_response.content)
        
        elif self.provider in ("anthropic", ProviderModel.ANTHROPIC):
            if response.stop_reason != 'tool_use':
                return []

            message: dict = {'role': 'user', 'content': []}
            for tool in response.content:
                if tool.type == "tool_use":
                    if stream:
                        tool = tool.model_dump()
                    if tool.name in self.local_tools.get_registered_tools():
                        result = self.local_tools.run_tools(tool)
                        message['content'].append(result.model_dump())
                    else:
                        run_tool_request = RunToolsRequest(
                            tool, self.provider, self.metadata)
                        run_response = self.tools.run_tools(run_tool_request)
                        output = run_response.content
                        message['content'].append(output)
            if message['content']:
                if append:
                    messages.append({'role': 'assistant', 'content': response.content})
                messages.append(message)

        else:
            raise NotImplementedError("Provider not supported")

        return messages

    @classmethod
    def _enum_matching(
        cls, value: Union[str, Enum], enum_values: List[str], variable_name: str
    ):
        str_value = value.value if isinstance(value, Enum) else value
        if str_value in enum_values:
            return value
        else:
            raise ValueError(
                f"Invalid value for {variable_name}: must match one of {enum_values}"
            )



from typing import Literal
from pydantic import BaseModel


class AnthropicToolResponse(BaseModel):
    """Represents the results of a tool call for Anthropic."""

    tool_use_id: str
    """The ID of the tool call."""

    content: str
    """Result of the tool call."""

    type: Literal["tool_result"]
    """The type of tool call the output is required for."""


class OpenAIToolResponse(BaseModel):
    """Represents the results of a tool call for OpenAI."""

    role: Literal['tool']

    tool_call_id: str
    """The ID of the tool call."""

    name: str
    """tool_function_name"""

    content: str
    """Result of the tool call."""



# This file was generated by liblab | https://liblab.com/

from .base import BaseModel
from enum import Enum


class Type_(Enum):
    TOOL_RESULT = "tool_result"

    def list():
        return list(map(lambda x: x.value, Type_._member_map_.values()))


class AnthropicToolCallResults(BaseModel):
    """
    Represents the results of a tool call for Anthropic.
    """

    def __init__(self, type: Type_, content: str, tool_use_id: str, **kwargs):
        """
        Initialize AnthropicToolCallResults
        Parameters:
        ----------
            type: str
            content: str
            tool_use_id: str
        """
        self.type = self._enum_matching(type, Type_.list(), "type")
        self.content = content
        self.tool_use_id = tool_use_id



# This file was generated by liblab | https://liblab.com/

from .base import BaseModel
from enum import Enum


class Role(Enum):
    TOOL = "tool"

    def list():
        return list(map(lambda x: x.value, Role._member_map_.values()))


class OpenAiToolCallResults(BaseModel):
    """
    Represents the results of a tool call for OpenAI.
    """

    def __init__(
        self, content: str, name: str, tool_call_id: str, role: Role, **kwargs
    ):
        """
        Initialize OpenAiToolCallResults
        Parameters:
        ----------
            content: str
            name: str
            tool_call_id: str
            role: str
        """
        self.content = content
        self.name = name
        self.tool_call_id = tool_call_id
        self.role = self._enum_matching(role, Role.list(), "role")



# This file was generated by liblab | https://liblab.com/
from __future__ import annotations
from .base import BaseModel
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .OpenAiFunction import OpenAiFunction


class Type_(Enum):
    """Function _Type"""
    FUNCTION = "function"

    @classmethod
    def list(cls):
        """Return the list of Items"""
        return [member.value for member in cls]


class OpenAiToolCall(BaseModel):
    """
    Represents a tool call for OpenAI.
    """

    def __init__(self, type: Type_, function: OpenAiFunction, id: str, **kwargs):
        """
        Initialize OpenAiToolCall
        Parameters:
        ----------
            type: str
            function: OpenAiFunction
            id: str
        """
        self.type = self._enum_matching(type, Type_.list(), "type")
        self.function = function
        self.id = id
    


# This file was generated by liblab | https://liblab.com/

from time import sleep
import requests
from http_exceptions import HTTPException, client_exceptions, server_exceptions
from json import JSONDecodeError
from .http_content_types import multipart_form_data_request
from .utils import to_serialize, rename_reserved_keys, rename_to_reserved_keys


class HTTPClient:
    """
    Provides functionality for invoking HTTP-based API calls to web services.
    """

    _retry_codes = [500, 503, 504]
    """list[int]: A list of status codes that invoke a retry."""

    _initial_delay = 150
    """int: The delay (in milliseconds) before performing a retry."""

    _max_retries = 3
    """int: The maximum number of retries."""

    def __init__(self, hook):
        self._hook = hook

    def _make_http_request(self, method, endpoint_url, headers, body_input):
        """
        Places API calls according to the HTTP method and content type.

        Parameters:
        ----------
            method : str
                The type of http call to perform
            endpoint_url : url
                The endpoint url to make the http call on
            headers : dict
                The http call's headers
            body_input : Any
                The request's body
        """
        request_method = getattr(requests, method)
        serialized_body = rename_to_reserved_keys(to_serialize(body_input))
        if "Content-Type" in headers:
            data_type, subtype = headers["Content-Type"].split("/")
            if data_type == "multipart":
                return multipart_form_data_request(
                    method, endpoint_url, headers, serialized_body
                )
            if data_type in ["text", "image"]:
                return request_method(
                    endpoint_url, headers=headers, data=serialized_body
                )
        if not serialized_body or method in {"get", "delete"}:
            return request_method(endpoint_url, headers=headers)
        return request_method(endpoint_url, headers=headers, json=serialized_body)

    def delete(self, endpoint_url: str, headers: dict, retry: bool = True):
        """
        Places API DELETE request and handles errors

        Parameters:
        ----------
            endpoint_url : str
                The endpoint url to make the http call on
            headers : dict
                The http call's headers
            retry : bool
                A boolean representing wether to attempt a retry
        """

        response = self._make_http_request("delete", endpoint_url, headers, None)
        if response.status_code in self._retry_codes and retry:
            try_cnt = 1
            while (
                response.status_code in self._retry_codes
                and try_cnt - 1 < self._max_retries
            ):
                sleep(self._initial_delay ** (try_cnt - 1) / 1000)
                response = self._make_http_request(
                    "delete", endpoint_url, headers, None
                )
                try_cnt += 1

        return self._handle_response(response)

    def get(self, endpoint_url: str, headers: dict, retry: bool = True):
        """
        Places an API GET request and handles errors

        Parameters:
        ----------
            endpoint_url : str
                The endpoint url to make the http call on
            headers : dict
                The http call's headers
            retry : bool
                A boolean representing wether to attempt a retry
        """

        response = self._make_http_request("get", endpoint_url, headers, None)
        if response.status_code in self._retry_codes and retry:
            try_cnt = 1
            while (
                response.status_code in self._retry_codes
                and try_cnt - 1 < self._max_retries
            ):
                sleep(self._initial_delay ** (try_cnt - 1) / 1000)
                response = self._make_http_request("get", endpoint_url, headers, None)
                try_cnt += 1

        return self._handle_response(response)

    def patch(self, endpoint_url: str, headers: dict, body_input, retry: bool = True):
        """
        Places an API PATCH request and handles errors

        Parameters:
        ----------
            endpoint_url : str
                The endpoint url to make the http call on
            headers : dict
                The http call's headers
            body_input:
                The patch request body
            retry : bool
                A boolean representing wether to attempt a retry
        """

        response = self._make_http_request("patch", endpoint_url, headers, body_input)
        if response.status_code in self._retry_codes and retry:
            try_cnt = 1
            while (
                response.status_code in self._retry_codes
                and try_cnt - 1 < self._max_retries
            ):
                sleep(self._initial_delay ** (try_cnt - 1) / 1000)
                response = self._make_http_request(
                    "patch", endpoint_url, headers, body_input
                )
                try_cnt += 1

        return self._handle_response(response)

    def post(self, endpoint_url: str, headers: dict, body_input, retry: bool = True):
        """
        Places an API POST request and handles errors

        Parameters:
        ----------
            endpoint_url : str
                The endpoint url to make the http call on
            headers : dict
                The http call's headers
            body_input:
                The post request body
            retry : bool
                A boolean representing wether to attempt a retry
        """

        response = self._make_http_request("post", endpoint_url, headers, body_input)
        if response.status_code in self._retry_codes and retry:
            try_cnt = 1
            while (
                response.status_code in self._retry_codes
                and try_cnt - 1 < self._max_retries
            ):
                sleep(self._initial_delay ** (try_cnt - 1) / 1000)
                response = self._make_http_request(
                    "post", endpoint_url, headers, body_input
                )
                try_cnt += 1

        return self._handle_response(response)

    def put(self, endpoint_url: str, headers: dict, body_input, retry: bool = True):
        """
        Places an API PUT request and handles errors

        Parameters:
        ----------
            endpoint_url : str
                The endpoint url to make the http call on
            headers : dict
                The http call's headers
            body_input:
                The put request body
            retry : bool
                A boolean representing whether to attempt a retry
        """

        response = self._make_http_request("put", endpoint_url, headers, body_input)
        if response.status_code in self._retry_codes and retry:
            try_cnt = 1
            while (
                response.status_code in self._retry_codes
                and try_cnt - 1 < self._max_retries
            ):
                sleep(self._initial_delay ** (try_cnt - 1) / 1000)
                response = self._make_http_request(
                    "put", endpoint_url, headers, body_input
                )
                try_cnt += 1

        return self._handle_response(response)

    def _create_exception_message(self, response, header: str) -> str:
        """
        Creates an exception message using a specific header

        Parameters:
        ----------
            response:
                Response data received from an http request
            header : str
                Header name for the created exception
        """
        if header in response.headers:
            return f"{response.text}, Headers {header}: {response.headers[header]}"
        return response.text

    def _handle_response(self, response: dict):
        """
        Handles a response from an API call

        Parameters:
        ----------
            response:
                Response data received from an http request
        """
        if response.status_code >= 200 and response.status_code < 400:
            try:
                return rename_reserved_keys(response.json())
            except JSONDecodeError:
                return response
        else:
            self._raise_from_status(response)

    def _raise_from_status(self, response) -> None:
        """
        Raises exception based response status, with additional information appended if useful

        Parameters:
        ----------
            response:
                Response data received from an http request
        """
        if response.status_code == 401:
            raise client_exceptions.UnauthorizedException(
                message=self._create_exception_message(response, "WWW-Authenticate")
            )
        elif response.status_code == 405:
            # this indicates a bug in the spec if it allows a method that the server rejects
            raise client_exceptions.MethodNotAllowedException(
                message=self._create_exception_message(response, "Allow")
            )
        elif response.status_code == 407:
            raise client_exceptions.ProxyAuthenticationRequiredException(
                message=self._create_exception_message(response, "Proxy-Authenticate")
            )
        elif response.status_code == 413:
            raise client_exceptions.PayloadTooLargeException(
                message=self._create_exception_message(response, "Retry-After")
            )
        elif response.status_code == 429:
            raise client_exceptions.TooManyRequestsException(
                message=self._create_exception_message(response, "Retry-After")
            )
        elif response.status_code == 503:
            raise server_exceptions.ServiceUnavailableException(
                message=self._create_exception_message(response, "Retry-After")
            )
        else:
            raise HTTPException.from_status_code(status_code=response.status_code)(
                message=response.text
            )



# This file was generated by liblab | https://liblab.com/

from __future__ import annotations
from .base import BaseModel
from typing import List
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .FunctionParameter import FunctionParameter


class Type_(Enum):
    OBJECT = "object"

    def list():
        return list(map(lambda x: x.value, Type_._member_map_.values()))


class Properties(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)


class FunctionParameters(BaseModel):
    """
    Tools Function Parameters
    """

    def __init__(
        self, type: Type_, required: List[str], properties: Properties, **kwargs
    ):
        """
        Initialize FunctionParameters
        Parameters:
        ----------
            type: str
            required: list of FunctionParametersRequired
            properties: Properties
        """
        self.type = self._enum_matching(type, Type_.list(), "type")
        self.required = required
        self.properties = properties



# This file was generated by liblab | https://liblab.com/

from .base import BaseModel
from typing import Any
from enum import Enum


class Type_(Enum):
    TOOL_USE = "tool_use"

    def list():
        return list(map(lambda x: x.value, Type_._member_map_.values()))


class AnthropicToolCall(BaseModel):
    """
    Represents a tool call for Anthropic.
    """

    def __init__(self, type: Type_, input: Any, name: str, id: str, **kwargs):
        """
        Initialize AnthropicToolCall
        Parameters:
        ----------
            type: str
            input
            name: str
            id: str
        """
        self.type = self._enum_matching(type, Type_.list(), "type")
        self.input = input
        self.name = name
        self.id = id



# This file was generated by liblab | https://liblab.com/

from __future__ import annotations
from .base import BaseModel
from typing import Any
from typing import Union
from .AnthropicToolCall import AnthropicToolCall
from .OpenAiAssistantsToolCall import OpenAiAssistantsToolCall
from .OpenAiToolCall import OpenAiToolCall
from .Provider import Provider


def returnAnthropicToolCall(input_data):
    return AnthropicToolCall(**input_data)


def returnOpenAiAssistantsToolCall(input_data):
    return OpenAiAssistantsToolCall(**input_data)


def returnOpenAiToolCall(input_data):
    return OpenAiToolCall(**input_data)


class ContentGuard(BaseModel):
    required_lists: dict = {
        "AnthropicToolCall": ["id", "input_", "name", "type"],
        "OpenAiAssistantsToolCall": ["id", "function", "type"],
        "OpenAiToolCall": ["id", "function"],
    }
    optional_lists: dict = {
        "AnthropicToolCall": [],
        "OpenAiAssistantsToolCall": [],
        "OpenAiToolCall": [],
    }
    class_list: dict = {
        "AnthropicToolCall": returnAnthropicToolCall,
        "OpenAiAssistantsToolCall": returnOpenAiAssistantsToolCall,
        "OpenAiToolCall": returnOpenAiToolCall,
    }

    @classmethod
    def return_one_of(cls, raw_input):
        return cls._one_of(
            cls.required_lists, cls.optional_lists, cls.class_list, raw_input
        )


Content = Union[AnthropicToolCall, OpenAiAssistantsToolCall, OpenAiToolCall]


class RunToolsRequest(BaseModel):
    """
    Represents a tool call for Toolhouse.
    """

    def __init__(self, content: Content, provider: Provider, metadata: Any, **kwargs):
        """
        Initialize RunToolsRequest
        Parameters:
        ----------
            content: Content
            provider: str
            metadata: dict
        """
        self.content = content
        self.provider = self._enum_matching(provider, Provider.list(), "provider")
        self.metadata = metadata



# This file was generated by liblab | https://liblab.com/

"""
Creates a BaseService class.
Performs API calls,sets authentication tokens and handles http exceptions.

Class:
    BaseService
"""
import platform
from typing import List, Union
from enum import Enum
import re
from ..net.http_client import HTTPClient


class BaseService:
    """
    A class to represent a base serivce

    Attributes
    ----------
    _url_prefix : str
        The base URL

    Methods
    -------
    set_access_token(token: str) -> None:
        Sets bearer token key
    def _add_required_headers(headers: dict):
        Request authorization headers
    def set_base_url(url: str):
        Sets the base url
    """

    _url_prefix = "http://api.example.com"

    _http = HTTPClient(None)

    def __init__(self, access_token: str) -> None:
        """
        Initialize client

        Parameters:
        ----------
           access_token : str
                A Access access token
        """
        self._access_token = access_token

    @classmethod
    def _pattern_matching(cls, value: str, pattern: str, variable_name: str):
        if re.match(r"{}".format(pattern), value):
            return value
        else:
            raise ValueError(f"Invalid value for {variable_name}: must match {pattern}")

    @classmethod
    def _enum_matching(
        cls, value: Union[str, Enum], enum_values: List[str], variable_name: str
    ):
        str_value = value.value if isinstance(value, Enum) else value
        if str_value in enum_values:
            return value
        else:
            raise ValueError(
                f"Invalid value for {variable_name}: must match one of {enum_values}"
            )

    def set_base_url(self, url: str) -> None:
        """
        Sets the base URL

        Parameters:
        ----------
            url:
                The base URL
        """
        self._url_prefix = url

    def set_access_token(self, token: str) -> None:
        """
        Sets access token key

        Parameters
        ----------
        token: string
            Access token value
        """
        self._access_token = token

    def _add_required_headers(self, headers: dict):
        """
        Request authorization headers

        Parameters
        ----------
        headers: dict
            Headers dict to add auth headers to
        """
        headers["User-Agent"] = f"Toolhouse/1.0.0 Python/{platform.python_version()}"
        headers["Authorization"] = f"Bearer {self._access_token}"
        return headers



# This file was generated by liblab | https://liblab.com/
from __future__ import annotations
from .base import BaseModel
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .OpenAiAssistantsFunction import OpenAiAssistantsFunction


class OpenAiAssistantsToolCall(BaseModel):
    """
    Represents a tool call for OpenAI.
    """

    def __init__(self, function: OpenAiAssistantsFunction, id: str, **kwargs):
        """
        Initialize OpenAiAssistantsToolCall
        Parameters:
        ----------
            function: OpenAiFunction
            id: str
        """
        self.function = function
        self.id = id



from typing import Union, List
from functools import wraps
import json
from anthropic.types import ToolUseBlock
from openai.types.chat import ChatCompletionMessageToolCall
from groq.types.chat import ChatCompletionMessageToolCall as GroqChatCompletionMessageToolCall
from ..models.RunLocalTools import AnthropicToolResponse, OpenAIToolResponse

SupportedTools = Union[
    ChatCompletionMessageToolCall,
    GroqChatCompletionMessageToolCall,
    ToolUseBlock]


class LocalTools():
    """Local Runner"""
    def __init__(self):
        self.local_tools = {}

    def get_registered_tools(self) -> List[str]:
        """Get List of registered tools"""
        return self.local_tools.keys()

    def register_local_tool(self, local_tool):
        """Register a new local tool runner"""
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                result = func(*args, **kwargs)
                if not isinstance(result, str):
                    raise TypeError(f"The function '{local_tool}' did not return a string. Returned type: {type(result).__name__}")
                return result
            # Add the function to the routes dictionary with the given name
            self.local_tools[local_tool] = wrapper
            return wrapper
        return decorator

    def _run_local_tools(self, local_tool, *args, **kwargs) -> str:
        """execute tool"""
        if local_tool in self.local_tools:
            return self.local_tools[local_tool](*args, **kwargs)
        else:
            raise ValueError(
                f"The LLM attempted to call a tool named {local_tool}, but there wasn't a registered local tool to respond to it."
                f"Make sure that you have a function that can handle this tool and decorate it with @register_local_tool(\"{local_tool}\")."
            )

    def run_tools(self, tool: SupportedTools) -> Union[AnthropicToolResponse, OpenAIToolResponse]:
        """_summary_

        Args:
            provider (Provider): _description_
            tool (_type_): _description_

        Returns:
            str: _description_
        """
        if isinstance(tool, ToolUseBlock):
            tool_input = tool.input if isinstance(tool.input, dict) else {}
            content = self._run_local_tools(local_tool=tool.name, **tool_input)
            return AnthropicToolResponse(type="tool_result", content=content, tool_use_id=tool.id)
        elif isinstance(tool, ChatCompletionMessageToolCall) or isinstance(tool, GroqChatCompletionMessageToolCall):
            content = self._run_local_tools(tool.function.name, **json.loads(tool.function.arguments))
            return OpenAIToolResponse(tool_call_id=tool.id, name=tool.function.name, content=content, role="tool")
        else:
            raise ValueError("Provider not suppoted")




# This file was generated by liblab | https://liblab.com/

"""
Collection of API calls according to the HTTP method and content type.

Functions:
    multipart_form_data_request
"""
import requests
import io
from mimetypes import guess_type


def multipart_form_data_request(method, endpoint_url, headers, body_input):
    """
    Places a multipart/formdata http request.

    Parameters:
    ----------
        method : str
            The type of http call to perform
        endpoint_url : url
            The endpoint url to make the http call on
        headers : dict
            The http call's headers
        body_input : Any
            The request's body
    """
    data = {}
    files = {}
    request_method = getattr(requests, method)
    del headers["Content-Type"]
    for key, value in body_input.items():
        if isinstance(value, (io.TextIOWrapper, io.BufferedIOBase)):
            mime_type, encoding = guess_type(value.name)
            file_tuple = (
                (value.name, value, mime_type) if mime_type else (value.name, value)
            )
            files[key] = file_tuple
        else:
            data[key] = value
    if files:
        return request_method(endpoint_url, headers=headers, files=files, data=data)
    return request_method(endpoint_url, headers=headers, data=data)



# This file was generated by liblab | https://liblab.com/

from .base import BaseModel


class OpenAiToolAssistantsCallResults(BaseModel):
    """
    Represents the results of a tool call for OpenAI Assistants.
    """

    def __init__(self, output: str, tool_call_id: str, **kwargs):
        """
        Initialize OpenAiToolAssistantsCallResults
        Parameters:
        ----------
            output: str
            tool_call_id: str
        """
        self.output = output
        self.tool_call_id = tool_call_id



# This file was generated by liblab | https://liblab.com/
# flake8: noqa
from .sdk import Toolhouse
from .exceptions import ToolhouseError
from .net.environment import Environment
from .models.Provider import Provider
from .models import stream_to_chat_completion, OpenAIStream
